import json
from json import JSONDecodeError
from subprocess import run
from typing import List, Optional

from .credentials import Credentials
from .entry import Entry


class Rbw:
    def list_entries(self) -> List[Entry]:
        rbw = run(["rbw", "list", "--fields", "folder,name,user"], encoding="utf-8", capture_output=True)

        if rbw.returncode != 0:
            print("There was a problem calling rbw. Is it correctly configured?")
            print(rbw.stderr)
            exit(2)

        return sorted(
            [self.__parse_rbw_output(it) for it in (rbw.stdout.strip("\n").split("\n"))],
            key=lambda x: x.folder.lower() + x.name.lower(),
        )

    def __parse_rbw_output(self, rbw_string: str) -> "Entry":
        fields = rbw_string.split("\t")

        try:
            return Entry(fields[1], fields[0], fields[2] if len(fields) > 2 else "")
        except IndexError:
            raise Exception("Entry is incorrectly formatted and cannot be parsed")

    def fetch_credentials(self, entry: Entry) -> "Credentials":
        try:
            data = json.loads(self.__load_from_rbw(entry.name, entry.username, entry.folder).strip())

            if data["data"] is None or "password" not in data["data"]:
                print("rofi-rbw only supports logins")
                return Credentials(
                    entry.name,
                    data["folder"],
                    entry.username,
                    notes=data["notes"],
                    further={item["name"]: item["value"] for item in data["fields"]},
                )

            return Credentials(
                entry.name,
                data["folder"],
                entry.username,
                data["data"]["password"] or "",
                data["data"]["totp"] is not None,
                data["notes"],
                [item["uri"] for item in data["data"]["uris"]],
                {item["name"]: item["value"] for item in data["fields"]},
            )

        except JSONDecodeError as exception:
            print(f"Could not parse the output: {exception.msg}")
            exit(12)

    def __load_from_rbw(self, name: str, username: str, folder: Optional[str]) -> str:
        command = ["rbw", "get", "--raw", name]
        if username:
            command.append(username)

        if folder:
            command.extend(["--folder", folder])

        return run(command, capture_output=True, encoding="utf-8").stdout

    def sync(self):
        run(["rbw", "sync"])



from dataclasses import dataclass
from enum import Enum
from typing import List, Union


class Action(Enum):
    TYPE = "type"
    COPY = "copy"
    PRINT = "print"
    SYNC = "sync"
    CANCEL = "cancel"


class Target:
    __value: str

    def __init__(self, input: str):
        prepared_input = input.strip()
        if prepared_input.lower() in (
            "username",
            "password",
            "menu",
            "totp",
            "notes",
        ) or prepared_input.lower().startswith("uri"):
            self.__value = prepared_input.lower()
        else:
            self.__value = prepared_input

    def is_uri(self) -> bool:
        return self.__value.startswith("uri")

    def uri_index(self) -> int:
        if not self.is_uri():
            raise IndexError()

        if len(self.__value) == 3:
            return 0
        else:
            return int(self.__value[4:]) - 1

    @property
    def raw(self):
        return self.__value

    def __eq__(self, other):
        return self.__value == other.__value


class Targets:
    USERNAME = Target("username")
    PASSWORD = Target("password")
    TOTP = Target("totp")
    NOTES = Target("notes")
    MENU = Target("menu")


@dataclass
class Keybinding:
    shortcut: str
    action: Action
    targets: Union[List[Target], None]



import time
from subprocess import run

from .abstractionhelper import is_installed, is_wayland


class Typer:
    @staticmethod
    def best_option(name: str = None) -> "Typer":
        try:
            return next(typer for typer in Typer.__subclasses__() if typer.name() == name)()
        except StopIteration:
            try:
                return next(typer for typer in Typer.__subclasses__() if typer.supported())()
            except StopIteration:
                return Typer()

    @staticmethod
    def supported() -> bool:
        pass

    @staticmethod
    def name() -> str:
        pass

    def get_active_window(self) -> str:
        raise NoTyperFoundException()

    def type_characters(self, characters: str, active_window: str) -> None:
        raise NoTyperFoundException()


class XDoToolTyper(Typer):
    @staticmethod
    def supported() -> bool:
        return not is_wayland() and is_installed("xdotool")

    @staticmethod
    def name() -> str:
        return "xdotool"

    def get_active_window(self) -> str:
        return run(args=["xdotool", "getactivewindow"], capture_output=True, encoding="utf-8").stdout[:-1]

    def type_characters(self, characters: str, active_window: str) -> None:
        run(
            [
                "xdotool",
                "windowactivate",
                "--sync",
                active_window,
                "type",
                "--clearmodifiers",
                "--delay",
                "0",
                characters,
            ]
        )
        # workaround for https://github.com/jordansissel/xdotool/issues/43
        run(["xdotool", "keyup", "Shift_L", "Shift_R", "Alt_L", "Alt_R"])


class WTypeTyper(Typer):
    @staticmethod
    def supported() -> bool:
        return is_wayland() and is_installed("wtype")

    @staticmethod
    def name() -> str:
        return "wtype"

    def get_active_window(self) -> str:
        return "not possible with wtype"

    def type_characters(self, characters: str, active_window: str) -> None:
        run(["wtype", characters])


class YDotoolTyper(Typer):
    @staticmethod
    def supported() -> bool:
        return is_wayland() and is_installed("ydotool")

    @staticmethod
    def name() -> str:
        return "ydotool"

    def get_active_window(self) -> str:
        return "not possible with ydotool"

    def type_characters(self, characters: str, active_window: str) -> None:
        time.sleep(0.05)
        run(["ydotool", "type", "--key-delay", "0", characters])


class DotoolTyper(Typer):
    @staticmethod
    def supported() -> bool:
        return is_installed("dotool")

    @staticmethod
    def name() -> str:
        return "dotool"

    def get_active_window(self) -> str:
        return "not possible with dotool"

    def type_characters(self, characters: str, active_window: str) -> None:
        run(["dotool"], text=True, input=f"type {characters}")


class NoTyperFoundException(Exception):
    def __str__(self) -> str:
        return "Could not find a valid way to type characters. Please check the required dependencies."



import argparse
import shlex
from typing import List, Tuple, Union

import configargparse

from . import __version__
from .cache import Cache
from .clipboarder import Clipboarder
from .credentials import Credentials
from .models import Action, Keybinding, Target, Targets
from .paths import *
from .rbw import Rbw
from .selector import Selector
from .typer import Typer


class RofiRbw(object):
    def __init__(self) -> None:
        self.args = self.__parse_arguments()
        self.rbw = Rbw()
        self.selector = Selector.best_option(self.args.selector)
        self.typer = Typer.best_option(self.args.typer)
        self.clipboarder = Clipboarder.best_option(self.args.clipboarder)
        self.active_window = self.typer.get_active_window()

    def __parse_arguments(self) -> argparse.Namespace:
        parser = configargparse.ArgumentParser(
            description="Insert or copy passwords and usernames from Bitwarden using rofi or rofi-likes.",
            default_config_files=config_file_locations,
        )
        parser.add_argument("--version", action="version", version="rofi-rbw " + __version__)
        parser.add_argument(
            "--action",
            "-a",
            dest="action",
            action="store",
            choices=[action.value for action in Action],
            default=Action.TYPE.value,
            help="What to do with the selected entry",
        )
        parser.add_argument(
            "--target", "-t", dest="targets", action="append", help="Which part of the entry do you want?"
        )
        parser.add_argument(
            "--prompt", "-r", dest="prompt", action="store", default="Choose entry", help="Set rofi-rbw's prompt"
        )
        parser.add_argument(
            "--selector-args",
            dest="selector_args",
            action="store",
            default="",
            help="A string of arguments to give to the selector",
        )
        parser.add_argument(
            "--selector",
            dest="selector",
            action="store",
            type=str,
            choices=["rofi", "wofi"],
            default=None,
            help="Choose the selector frontend",
        )
        parser.add_argument(
            "--clipboarder",
            dest="clipboarder",
            action="store",
            type=str,
            choices=["xsel", "xclip", "wl-copy"],
            default=None,
            help="Choose the application to access the clipboard with",
        )
        parser.add_argument(
            "--typer",
            dest="typer",
            action="store",
            type=str,
            choices=["xdotool", "wtype", "ydotool", "dotool"],
            default=None,
            help="Choose the application to type with",
        )
        parser.add_argument(
            "--clear-after",
            dest="clear",
            action="store",
            type=int,
            default=0,
            help="Limit the duration in seconds passwords stay in your clipboard. When not set or <= 0, passwords stay indefinitely.",
        )
        parser.add_argument(
            "--no-help", dest="show_help", action="store_false", help="Don't show a help message about the shortcuts"
        )
        parser.add_argument(
            "--no-folder", dest="show_folders", action="store_false", help="Don't show the entry's folder"
        )
        parser.set_defaults(show_folders=True)
        parser.add_argument("--no-cache", dest="use_cache", action="store_false", help="Don't save history in cache")
        parser.set_defaults(use_cache=True)
        parser.add_argument(
            "--keybindings",
            dest="keybindings",
            action="store",
            type=str,
            default=",".join(
                [
                    "Alt+1:type:username:password",
                    "Alt+2:type:username",
                    "Alt+3:type:password",
                    "Alt+4:type:totp",
                    "Alt+c:copy:password",
                    "Alt+u:copy:username",
                    "Alt+t:copy:totp",
                    "Alt+m::menu",
                    "Alt+s:sync",
                ]
            ),
            help="Define keyboard shortcuts in the format <shortcut>:<action>:<target>, separated with a comma",
        )
        parser.add_argument(
            "--menu-keybindings",
            dest="menu_keybindings",
            action="store",
            type=str,
            default=",".join(["Alt+t:type", "Alt+c:copy"]),
            help="Define the keyboard shortcuts in the target menu in the format <shortcut>:<action> separated with a comma",
        )

        parsed_args = parser.parse_args()

        parsed_args.selector_args = shlex.split(parsed_args.selector_args)

        parsed_args.action = Action(parsed_args.action)

        if parsed_args.targets:
            parsed_args.targets = [Target(target) for target in parsed_args.targets]
        else:
            parsed_args.targets = [Targets.USERNAME, Targets.PASSWORD]

        parsed_args.parsed_keybindings = []
        for keybinding in parsed_args.keybindings.split(","):
            elements = keybinding.split(":")
            parsed_args.parsed_keybindings.append(
                Keybinding(
                    elements[0],
                    Action(elements[1]) if elements[1] else None,
                    [Target(target_string) for target_string in elements[2:]],
                )
            )

        parsed_args.parsed_menu_keybindings = []
        for keybinding in parsed_args.menu_keybindings.split(","):
            elements = keybinding.split(":")
            parsed_args.parsed_menu_keybindings.append(
                Keybinding(elements[0], Action(elements[1]) if elements[1] else None, None)
            )

        return parsed_args

    def main(self) -> None:
        entries = self.rbw.list_entries()

        if self.args.use_cache:
            cache = Cache()
            entries = cache.sorted(entries)

        (selected_targets, selected_action, selected_entry) = self.selector.show_selection(
            entries,
            self.args.prompt,
            self.args.show_help,
            self.args.show_folders,
            self.args.parsed_keybindings,
            self.args.selector_args,
        )

        if selected_action == Action.SYNC:
            self.rbw.sync()
            (selected_targets, selected_action, selected_entry) = self.selector.show_selection(
                self.rbw.list_entries(),
                self.args.prompt,
                self.args.show_help,
                self.args.show_folders,
                self.args.parsed_keybindings,
                self.args.selector_args,
            )

        if selected_action == Action.CANCEL:
            return

        credential = self.rbw.fetch_credentials(selected_entry)

        if self.args.use_cache:
            cache.update(credential)

        if selected_targets is not None:
            self.args.targets = selected_targets

        if selected_action is not None:
            self.args.action = selected_action

        if Targets.MENU in self.args.targets:
            targets, action = self.__show_target_menu(
                credential,
                self.args.show_help,
            )
            self.args.targets = targets
            if action is not None:
                self.args.action = action

        self.__execute_action(credential)

    def __show_target_menu(
        self, cred: Credentials, show_help_message: bool
    ) -> Tuple[List[Target], Union[Action, None]]:
        targets, action = self.selector.select_target(
            cred, show_help_message, self.args.parsed_menu_keybindings, additional_args=self.args.selector_args
        )

        if action == Action.CANCEL:
            self.main()
            exit(0)

        return targets, action

    def __execute_action(self, cred: Credentials) -> None:
        if self.args.action == Action.TYPE:
            characters = "\t".join([cred[target] for target in self.args.targets])
            self.typer.type_characters(characters, self.active_window)
            if Targets.PASSWORD in self.args.targets and cred.totp != "":
                self.clipboarder.copy_to_clipboard(cred.totp)
        elif self.args.action == Action.COPY:
            for target in self.args.targets:
                self.clipboarder.copy_to_clipboard(cred[target])
            if len(self.args.targets) == 1 and self.args.targets[0] == Targets.PASSWORD:
                self.clipboarder.clear_clipboard_after(self.args.clear)
        elif self.args.action == Action.PRINT:
            print("\n".join([cred[target] for target in self.args.targets]))



import time
from subprocess import run

from .abstractionhelper import is_installed, is_wayland


class Clipboarder:
    @staticmethod
    def best_option(name: str = None) -> "Clipboarder":
        try:
            return next(clipboarder for clipboarder in Clipboarder.__subclasses__() if clipboarder.name() == name)()
        except StopIteration:
            try:
                return next(clipboarder for clipboarder in Clipboarder.__subclasses__() if clipboarder.supported())()
            except StopIteration:
                return Clipboarder()

    @staticmethod
    def supported() -> bool:
        pass

    @staticmethod
    def name() -> str:
        pass

    def copy_to_clipboard(self, characters: str) -> None:
        raise NoClipboarderFoundException()

    def clear_clipboard_after(self, clear: int) -> None:
        raise NoClipboarderFoundException()


class XSelClipboarder(Clipboarder):
    __last_copied_characters: str

    @staticmethod
    def supported() -> bool:
        return not is_wayland() and is_installed("xsel")

    @staticmethod
    def name() -> str:
        return "xsel"

    def copy_to_clipboard(self, characters: str) -> None:
        run(["xsel", "--input", "--clipboard"], input=characters, encoding="utf-8")

        self.__last_copied_characters = characters

    def fetch_clipboard_content(self) -> str:
        return run(
            [
                "xsel",
                "--output",
                "--clipboard",
            ],
            capture_output=True,
            encoding="utf-8",
        ).stdout

    def clear_clipboard_after(self, clear: int) -> None:
        if clear > 0:
            time.sleep(clear)

            # Only clear clipboard if nothing has been copied since the password
            if self.fetch_clipboard_content() == self.__last_copied_characters:
                run(["xsel", "--clear", "--clipboard"])
                self.__last_copied_characters = None


class XClipClipboarder(Clipboarder):
    __last_copied_characters: str

    @staticmethod
    def supported() -> bool:
        return not is_wayland() and is_installed("xclip")

    @staticmethod
    def name() -> str:
        return "xclip"

    def copy_to_clipboard(self, characters: str) -> None:
        run(["xclip", "-in", "-selection", "clipboard"], input=characters, encoding="utf-8")

        self.__last_copied_characters = characters

    def fetch_clipboard_content(self) -> str:
        return run(["xclip", "-o", "-selection", "clipboard"], capture_output=True, encoding="utf-8").stdout

    def clear_clipboard_after(self, clear: int) -> None:
        if clear > 0:
            time.sleep(clear)

            # Only clear clipboard if nothing has been copied since the password
            if self.fetch_clipboard_content() == self.__last_copied_characters:
                self.copy_to_clipboard("")
                self.__last_copied_characters = None


class WlClipboarder(Clipboarder):
    @staticmethod
    def supported() -> bool:
        return is_wayland() and is_installed("wl-copy")

    @staticmethod
    def name() -> str:
        return "wl-copy"

    def copy_to_clipboard(self, characters: str) -> None:
        run(["wl-copy"], input=characters, encoding="utf-8")

    def clear_clipboard_after(self, clear: int) -> None:
        if clear > 0:
            time.sleep(clear)
            run(["wl-copy", "--clear"])


class NoClipboarderFoundException(Exception):
    def __str__(self) -> str:
        return "Could not find a valid way to copy to clipboard. Please check the required dependencies."



import re
from subprocess import run
from typing import Dict, List, Tuple, Union

from .abstractionhelper import is_installed, is_wayland
from .credentials import Credentials
from .entry import Entry
from .models import Action, Keybinding, Target, Targets


class Selector:
    @staticmethod
    def best_option(name: str = None) -> "Selector":
        try:
            return next(selector for selector in Selector.__subclasses__() if selector.name() == name)()
        except StopIteration:
            try:
                return next(selector for selector in Selector.__subclasses__() if selector.supported())()
            except StopIteration:
                return Selector()

    @staticmethod
    def supported() -> bool:
        pass

    @staticmethod
    def name() -> str:
        pass

    def show_selection(
        self,
        entries: List[Entry],
        prompt: str,
        show_help_message: bool,
        show_folders: bool,
        keybindings: Dict[str, Tuple[Action, List[Target]]],
        additional_args: List[str],
    ) -> Tuple[Union[List[Target], None], Union[Action, None], Union[Entry, None]]:
        raise NoSelectorFoundException()

    def select_target(
        self,
        credentials: Credentials,
        show_help_message: bool,
        keybindings: Dict[str, Action],
        additional_args: List[str],
    ) -> Tuple[Union[List[Target], None], Union[Action, None]]:
        raise NoSelectorFoundException()

    def _format_targets_from_credential(self, credentials: Credentials) -> List[str]:
        targets = []
        if credentials.username:
            targets.append(f"Username: {credentials.username}")
        if credentials.password:
            targets.append(f'Password: {credentials.password[0]}{"*" * (len(credentials.password) - 1)}')
        if credentials.has_totp:
            targets.append(f"TOTP: {credentials.totp}")
        if credentials.notes:
            targets.append(f"Notes: {credentials.notes}")
        if len(credentials.uris) == 1:
            targets.append(f"URI: {credentials.uris[0]}")
        else:
            for key, value in enumerate(credentials.uris):
                targets.append(f"URI {key + 1}: {value}")
        for (key, value) in credentials.further.items():
            targets.append(f'{self._format_further_item_name(key)}: {value[0]}{"*" * (len(value) - 1)}')

        return targets

    def _format_further_item_name(self, key: str) -> str:
        if key.lower() in ["username", "password", "totp"] or re.match(r"^URI \d+$", key):
            return f"{key} (field)"
        return key

    @staticmethod
    def _extract_targets(output: str) -> List[Target]:
        return [Target(line.split(":")[0]) for line in output.strip().split("\n")]

    @staticmethod
    def _calculate_max_width(entries: List[Entry], show_folders: bool) -> int:
        if show_folders:
            return max(len(it.name) + len(it.folder) + 1 for it in entries)
        else:
            return max(len(it.name) for it in entries)

    @staticmethod
    def _format_folder(entry: Entry, show_folders: bool) -> str:
        if not show_folders or not entry.folder:
            return ""
        return f"{entry.folder}/"

    @staticmethod
    def justify(entry: Entry, max_width: int, show_folders: bool) -> str:
        whitespace_length = max_width - len(entry.name)
        if show_folders:
            whitespace_length -= len(entry.folder)
            if entry.folder:
                whitespace_length -= 1
        return " " * whitespace_length


class Rofi(Selector):
    @staticmethod
    def supported() -> bool:
        return is_installed("rofi")

    @staticmethod
    def name() -> str:
        return "rofi"

    def show_selection(
        self,
        entries: List[Entry],
        prompt: str,
        show_help_message: bool,
        show_folders: bool,
        keybindings: List[Keybinding],
        additional_args: List[str],
    ) -> Tuple[Union[List[Target], None], Union[Action, None], Union[Entry, None]]:
        parameters = [
            "rofi",
            "-markup-rows",
            "-dmenu",
            "-i",
            "-sort",
            "-p",
            prompt,
            *self.__build_parameters_for_keybindings(keybindings),
            *additional_args,
        ]

        if show_help_message and keybindings:
            parameters.extend(self.__format_keybindings_message(keybindings))

        rofi = run(
            parameters,
            input="\n".join(self.__format_entries(entries, show_folders)),
            capture_output=True,
            encoding="utf-8",
        )

        if rofi.returncode == 1:
            return None, Action.CANCEL, None
        elif rofi.returncode >= 10:
            keybinding = keybindings[(rofi.returncode - 10)]
            return_action = keybinding.action
            return_targets = keybinding.targets
        else:
            return_action = None
            return_targets = None

        return return_targets, return_action, self.__parse_formatted_string(rofi.stdout)

    def __format_entries(self, entries: List[Entry], show_folders: bool) -> List[str]:
        max_width = self._calculate_max_width(entries, show_folders)
        return [
            f"{self._format_folder(it, show_folders)}<b>{it.name}</b>{self.justify(it, max_width, show_folders)}  {it.username}"
            for it in entries
        ]

    def __parse_formatted_string(self, formatted_string: str) -> Entry:
        match = re.compile("(?:(?P<folder>.+)/)?<b>(?P<name>.*?) *</b>(?P<username>.*)").search(formatted_string)

        return Entry(match.group("name"), match.group("folder"), match.group("username").strip())

    def select_target(
        self,
        credentials: Credentials,
        show_help_message: bool,
        keybindings: List[Keybinding],
        additional_args: List[str],
    ) -> Tuple[Union[List[Target], None], Union[Action, None]]:
        parameters = [
            "rofi",
            "-markup-rows",
            "-dmenu",
            "-p",
            "Choose target",
            "-i",
            *self.__build_parameters_for_keybindings(keybindings),
            *additional_args,
        ]

        if show_help_message and keybindings:
            parameters.extend(self.__format_keybindings_message(keybindings))

        rofi = run(
            parameters,
            input="\n".join(self._format_targets_from_credential(credentials)),
            capture_output=True,
            encoding="utf-8",
        )

        if rofi.returncode == 1:
            return None, Action.CANCEL
        elif rofi.returncode >= 10:
            action = keybindings[(rofi.returncode - 10)].action
        else:
            action = None

        return (self._extract_targets(rofi.stdout)), action

    def __build_parameters_for_keybindings(self, keybindings: List[Keybinding]) -> List[str]:
        params = []
        for index, keybinding in enumerate(keybindings):
            params.extend([f"-kb-custom-{1 + index}", keybinding.shortcut])

        return params

    def __format_keybindings_message(self, keybindings: List[Keybinding]):
        return [
            "-mesg",
            " | ".join(
                [
                    f"<b>{keybinding.shortcut}</b>: {self.__format_action_and_targets(keybinding)}"
                    for keybinding in keybindings
                ]
            ),
        ]

    def __format_action_and_targets(self, keybinding: Keybinding) -> str:
        if keybinding.targets and Targets.MENU in keybinding.targets:
            return "Menu"
        elif keybinding.action == Action.SYNC:
            return "Sync logins"
        elif keybinding.targets:
            return f"{keybinding.action.value.title()} {', '.join([target.raw for target in keybinding.targets])}"
        else:
            return keybinding.action.value.title()


class Wofi(Selector):
    @staticmethod
    def supported() -> bool:
        return is_wayland() and is_installed("wofi")

    @staticmethod
    def name() -> str:
        return "wofi"

    def show_selection(
        self,
        entries: List[Entry],
        prompt: str,
        show_help_message: bool,
        show_folders: bool,
        keybindings: Dict[str, Tuple[Action, List[Target]]],
        additional_args: List[str],
    ) -> Tuple[Union[List[Target], None], Union[Action, None], Union[Entry, None]]:
        parameters = ["wofi", "--dmenu", "-p", prompt, *additional_args]

        wofi = run(
            parameters,
            input="\n".join(self.__format_entries(entries, show_folders)),
            capture_output=True,
            encoding="utf-8",
        )
        if wofi.returncode == 0:
            return None, None, self.__parse_formatted_string(wofi.stdout)
        else:
            return None, Action.CANCEL, None

    def __format_entries(self, entries: List[Entry], show_folders: bool) -> List[str]:
        max_width = self._calculate_max_width(entries, show_folders)
        return [
            f"{self._format_folder(it, show_folders)}{it.name}{self.justify(it, max_width, show_folders)}  {it.username}"
            for it in entries
        ]

    def __parse_formatted_string(self, formatted_string: str) -> Entry:
        match = re.compile("(?:(?P<folder>.+)/)?(?P<name>.*?) *  (?P<username>.*)").search(formatted_string)

        return Entry(match.group("name").strip(), match.group("folder"), match.group("username").strip())

    def select_target(
        self,
        credentials: Credentials,
        show_help_message: bool,
        keybindings: Dict[str, Action],
        additional_args: List[str],
    ) -> Tuple[Union[List[Target], None], Union[Action, None]]:
        parameters = ["wofi", "--dmenu", "-p", "Choose target", *additional_args]

        wofi = run(
            parameters,
            input="\n".join(self._format_targets_from_credential(credentials)),
            capture_output=True,
            encoding="utf-8",
        )

        if wofi.returncode == 1:
            return None, Action.CANCEL

        return self._extract_targets(wofi.stdout), None


class NoSelectorFoundException(Exception):
    def __str__(self) -> str:
        return "Could not find a valid way to show the selection. Please check the required dependencies."



from dataclasses import dataclass, field
from subprocess import run
from typing import Dict, List, Optional, Union

from .entry import Entry
from .models import Target, Targets


@dataclass(frozen=True)
class Credentials(Entry):
    password: Optional[str] = ""
    has_totp: bool = False
    notes: Optional[str] = ""
    uris: List[str] = field(default_factory=list)
    further: Dict[str, str] = field(default_factory=dict)

    def __getitem__(self, target: Target) -> Optional[Union[str, List[str]]]:
        if target == Targets.USERNAME:
            return self.username
        elif target == Targets.PASSWORD:
            return self.password
        elif target == Targets.TOTP:
            return self.totp
        elif target == Targets.NOTES:
            return self.notes
        elif target.is_uri():
            return self.uris[target.uri_index()]
        else:
            return self.further.get(target.raw.removesuffix(" (field)"), None)

    @property
    def totp(self):
        if not self.has_totp:
            return ""

        command = ["rbw", "code", self.name]
        if self.username:
            command.extend([self.username])
        if self.folder:
            command.extend(["--folder", self.folder])
        return run(command, capture_output=True, encoding="utf-8").stdout.strip()
